/**
 * AndropodInterface - high level adb / serial communication
 *
 * @author Bernhard Wrndl-Aichriedler <bwa@xdevelop.at>, Julian Nischler  <julian@xdevelop.at>
 * @copyright Copyright 2011, Upper Austria University of Applied Sciences & \
   Julian Nischler & Bernhard Wrndl-Aichriedler
 */

#include "AdbCommunication.h"

#include "vos.h"
#include "ioctl.h"
#include "USB.h"
#include "USBHost.h"

#include "Definitions.h"
#include "Debug.h"

#include "Adb.h"
#include "AdbSlave.h"
#include "USBHostGeneric.h"
#include "HardwareControl.h"


/**
 * Handle of the USB host
 */
VOS_HANDLE  adbUsbHostHandle 	= 0;

/**
 * Id of the USB Host
 */
uint8 		adbUsbHostId 		= 0;

/**
 *Handle of the generic USB
 */
VOS_HANDLE  adbUsbGenericHandle	= 0;

/**
 * Id of the generic USB driver
 */
uint8 		adbUsbGenericId 	= 0;

/**
 * Handle for the UART
 */
VOS_HANDLE  adbUartHandle		= 0;

/**
 * Handle of the USB host
 */
VOS_HANDLE  adbUsbSlaveHandle 	= 0;

/**
 * Id of the USB Host
 */
uint8 		adbUsbSlaveId 		= 0;

/**
 * Handle of the USB host
 */
VOS_HANDLE  adbUsbSlaveAdbHandle = 0;

/**
 * Id of the USB Host
 */
uint8 		adbUsbSlaveAdbId 	= 0;

/**
 * Mutex used for locking the reader thread
 */
vos_mutex_t	adbDevicePresent;


/**
 * Write thread control structure
 */
vos_tcb_t *adbWriterTcb;

/**
 * Uart reader thread control structure
 */
vos_tcb_t *adbUartReaderTcb;

/**
 * Usb reader thread control structure
 */
vos_tcb_t *adbUsbReaderTcb;


vos_mutex_t	adbHostWriteMutex;
vos_mutex_t	adbSlaveWriteMutex;
vos_mutex_t	adbHostReadMutex;
vos_mutex_t	adbSlaveReadMutex;


/**
 * Initialize the USB Host for adb communication
 */
void AdbCommunicationInit(uint8 usbHostId,
	uint8 genericUsbHostId,
	uint8 usbSlaveId,
	uint8 usbSlaveAdbId,
	VOS_HANDLE uartHandle)
{
	//Save variables
	//adbUsbHostHandle 		= usbHost;
	adbUsbHostId			= usbHostId;
	//adbUsbGenericHandle 	= genericUsbHost;
	adbUsbGenericId			= genericUsbHostId;
	adbUartHandle			= uartHandle;
	//adbUsbSlaveHandle		= usbSlave;
	adbUsbSlaveId			= usbSlaveId;
	//adbUsbSlaveAdbHandle	= usbSlaveAdb;
	adbUsbSlaveAdbId		= usbSlaveAdbId;

	//Init read locking mutex
	vos_init_mutex(&adbDevicePresent, 1);

	vos_init_mutex(&adbSlaveReadMutex, 0);
	vos_init_mutex(&adbSlaveWriteMutex, 0);

	vos_init_mutex(&adbHostReadMutex, 0);
	vos_init_mutex(&adbHostWriteMutex, 0);


	//Init Usb host
	AdbHostInit();

	// Initialise USB Slave
	AdbSlaveInit(&adbUsbSlaveHandle, adbUsbSlaveId,
		&adbUsbSlaveAdbHandle, adbUsbSlaveAdbId,
//		&adbSlaveReadMutex, &adbSlaveWriteMutex
//		&adbHostWriteMutex, &adbHostReadMutex
		&adbHostWriteMutex, &adbHostWriteMutex
	);

	//Initialize mutex of ADB Interface
	AdbInit();

	//System threads
	adbWriterTcb		= vos_create_thread_ex(20, 0x400, AdbThreadWriter, "Adb Writer",0);
	adbUartReaderTcb 	= vos_create_thread_ex(20, 0x400, AdbUartThreadReader, "Adb Uart Reader",0);
	adbUsbReaderTcb 	= vos_create_thread_ex(20, 0x400, AdbUsbThreadReader, "Adb Usb Reader",0);
}

/**
 * Initialize usb host and generic interface on top of it
 * @param 	usbHost				USB host handle
 * @param	usbHostId			USB host id
 * @param	genericUsbHost		Generic USB host handle
 * @param	genericUsbHostId	Generic USB host id
 */
void AdbHostInit()
{
	// USB Host configuration context
	usbhost_context_t usbhostContext;

	// Initialise USB Host
	usbhostContext.if_count = 4;
	usbhostContext.ep_count = 8;
	usbhostContext.xfer_count = 16;
	usbhostContext.iso_xfer_count = 0;
	usbhost_init(adbUsbHostId, -1, &usbhostContext);

	//Intialize generic usb driver
	usbhostGeneric_init(adbUsbGenericId,
//		&adbHostReadMutex, &adbHostWriteMutex
		&adbHostWriteMutex, &adbHostWriteMutex
	);

	//Open USB Generic driver
	adbUsbGenericHandle 	= vos_dev_open(adbUsbGenericId);
}

/**
 * Wait for a device to be attached and connect to it
 */
bool AdbWaitAndConnect()
{
	//Define datatypes
	uint8 i = 0;
	usbhostGeneric_ioctl_t generic_iocb;
	usbhostGeneric_ioctl_cb_attach_t genericAtt;
	usbhost_device_handle *ifDev;
	usbhost_ioctl_cb_t hc_iocb;
	usbhost_ioctl_cb_vid_pid_t hc_iocb_vid_pid;
	usbhost_ioctl_cb_class_t hc_iocb_class;

	//Set activity LED
	setActivityLed(false);

	//Open host driver
	adbUsbHostHandle = vos_dev_open(adbUsbHostId);

	//Set wanted device classes and protocols
	hc_iocb_class.dev_class        = USB_CLASS_ANDROID;
	hc_iocb_class.dev_subclass     = USB_SUBCLASS_ANDROID_ADB;
	hc_iocb_class.dev_protocol     = USB_PROTOCOL_ANDROID_DEFAULT;

	vos_delay_msecs(1000);

	//Loop until device connected successfully
	while(true)
	{
		//Print something to indicate activity
		debugPrint("*");

		//Use ioctl to see if bus available
		hc_iocb.ioctl_code =  VOS_IOCTL_USBHOST_GET_CONNECT_STATE;
		hc_iocb.get = &i;
		vos_dev_ioctl(adbUsbHostHandle, &hc_iocb);

		//If a device is connected (port enumerated)
		if (i == PORT_STATE_ENUMERATED)
		{
			//We found a device
			//debugPrint("\r\nBus Ready\r\n");

			//Delay a bit for the archos tablet
			vos_delay_msecs(1000);

			//Use ioctl to find first hub device
			hc_iocb.ioctl_code = VOS_IOCTL_USBHOST_DEVICE_FIND_HANDLE_BY_CLASS;
			hc_iocb.handle.dif = NULL;
			hc_iocb.set = &hc_iocb_class;
			hc_iocb.get = &ifDev;

			//If a device with the desired Interfaces has been found
			if (vos_dev_ioctl(adbUsbHostHandle, &hc_iocb) == USBHOST_OK)
			{
				//We found a device
				debugPrint("Found ADB\r\n");

				//Open USB Generic driver
				//adbUsbGenericHandle 	= vos_dev_open(adbUsbGenericId);
				genericAtt.hc_handle 	= adbUsbHostHandle;
				genericAtt.ifDev 		= ifDev;
				generic_iocb.ioctl_code = VOS_IOCTL_USBHOSTGENERIC_ATTACH;
				generic_iocb.set.att 	= &genericAtt;

				//Check if driver can be opened
				if (vos_dev_ioctl(adbUsbGenericHandle, &generic_iocb) == USBHOSTGENERIC_OK)
				{
					//Mark device as present -> start reader
					vos_unlock_mutex(&adbDevicePresent);

					if (AdbSendConnect(adbUsbGenericHandle) == 1)
					{
						//Set activity led
						setActivityLed(true);

						//Timeout counter
						i = 0;

						//Loop until connected
						while(!AdbIsConnected())
						{
							debugPrint("-");

							//If a timeout occured -> reset the system
							if(++i >= ADB_COMMUNICATION_CONNECT_TIMEOUT)
							{
								errorPrint("Timeout reached\r\n");
								/*
								 * The device is not read and doesn't respond
								 * with a connect within the timeout, reset
								 */
								AdbClose();
							}

							vos_delay_msecs(100);
						}

						debugPrint("\r\n");

						//Return from loop
						break;
					}
					else
					{
						//Stop reader again as send connect failed
						vos_trylock_mutex(&adbDevicePresent);
					}
				}

				//Close the generic driver properly
				vos_dev_close(adbUsbGenericHandle);
			}
			else
			{
				/*
				 * Unable to attach serial device,
				 * this generally happens with the Archos Internet Tablet
				 * as it's USB slave is simply too slow
				 */
				AdbClose();
			}
		}

		//Wait 1sec
		vos_delay_msecs(1000);
	}

	return true;
}

/**
 * Open a tcp connection tunnel to the attached and connected device
 */
bool AdbWaitAndOpenTcp()
{
	int timeoutCounter = 0;

	//Check if the connection has been established
	while(AdbIsConnected())
	{
		//Open the output port
		if (AdbSendOpen(adbUsbGenericHandle) == 1)
		{
			timeoutCounter = 0;

			//Wait until port is open
			while(!AdbIsOpen())
			{
				//If a timeout occured -> reset the system
				if(timeoutCounter >= ADB_COMMUNICATION_CONNECT_TIMEOUT)
				{
					//Timeout reached, no error just software on device
					//not started yet ;-)
					debugPrint("Timeout reached\r\n");

					//Break out and try to connect to app again
					break;
				}

				//Do some waiting and blinking
				debugPrint("x");
				setActivityLed(false);
				vos_delay_msecs(500);
				setActivityLed(true);
				vos_delay_msecs(500);

				timeoutCounter+=10; //1 sec clock
			}

			debugPrint("\r\n");
			//debugPrintInt(AdbIsConnected());
			debugPrint("\r\nEnded\r\n");

			//If no timeout occured -> return
			if(AdbIsOpen())
			{
				debugPrint("\r\n");
				return true;
			}
		}
	}

	debugPrint("Closing here??\r\n");

	//Something bad happended, close connection and reset
	AdbClose();

	return false;
}

/**
 * Close the adb connection and free resources
 */
bool AdbClose()
{
	/* We close it hard here */
	systemReset();

	/*
	//Print debug message
	debugPrint("Close adb\r\n");

	//Set activity led
	setActivityLed(false);

	//Device not present, lock reading thread
	vos_trylock_mutex(&adbDevicePresent);

	debugPrint("trylock mutex done!!!!\r\n");

	//If connection has been openend
	if(AdbIsOpen())
	{
		//Send adb close command
		AdbSendClose(adbUsbGenericHandle);
		debugPrint("AdbSendClose done!!!!\r\n");
	}

	//Close generic usb handle
	vos_dev_close(adbUsbGenericHandle);

	debugPrint("Closing completely done!!!!\r\n");

	//Reset to restart connection -> much more reliable
	//It's hard but to 100% reliable
	systemReset();

	*/
	return true;
}

/**
 * Writer thread handling the communication from Uart -> Adb
 */
void AdbThreadWriter()
{
	//Buffer
	char adbReaddata[10];
	unsigned short numToRead = 0;
	unsigned short numRead = 0;

	common_ioctl_cb_t uart_iocb;


	//Print that reader started
	debugPrint("AdbThreadWriter\r\n");

	while(true)
	{
		//Connect and wait until connected
		AdbWaitAndConnect();

		//If we are able to open a tcp connection
		if(AdbWaitAndOpenTcp()) //Wait for tcp socket to be available
		{
			//Read bytes from serial port, until error
			while(AdbIsOpen())
			{
				uart_iocb.ioctl_code = VOS_IOCTL_COMMON_GET_RX_QUEUE_STATUS;
				vos_dev_ioctl(adbUartHandle, &uart_iocb);

				numToRead = uart_iocb.get.queue_stat>10?10:uart_iocb.get.queue_stat;

				if(numToRead > 0)
				{
					if(vos_dev_read(adbUartHandle, (void *)adbReaddata, numToRead, &numRead) != 0)
					{
						break;
					}
					else
					{
						//Write to the adb interface
						if(!AdbSendWrite(adbUsbGenericHandle, adbReaddata, numRead))
						{
							//Unable to submit a packet -> close the port
							errorPrint("AdbSendWrite failed\r\n");
							break;
						}
						numRead = 0;
					}
				}
				else
				{
					vos_delay_msecs(50);
				}
			}
		}

		debugPrint("Closing port\r\n");

		//Close connection
		AdbClose();
	}
}

/**
 * Reader thread handling the communication from Adb -> Uart
 */
void AdbUartThreadReader()
{
	//Our debug reader thread started
	debugPrint("AdbUartThreadReader\r\n");

	while(true)
	{

		//Lock mutex and until device is present, and unlock it again
		vos_lock_mutex(&adbDevicePresent);
		vos_unlock_mutex(&adbDevicePresent);

		//Call adb package router and tell it to route to the Uart handle
		if(!AdbRouteMessage(adbUsbGenericHandle, adbUartHandle, adbUsbSlaveAdbHandle))
		{
			debugPrint("Route Message failed!");
			//Close connection if there has been an error
			AdbClose();
		}
	}
}


/**
 * Reader thread handling the communication from Adb Slave -> Adb Host
 */
void AdbUsbThreadReader()
{
	//Our debug reader thread started
	debugPrint("AdbUsbThreadReader\r\n");
	errorPrint("Starting\r\n");

	while(true)
	{
		if(adbUsbSlaveAdbHandle != NULL &&
			AdbSlaveAttached(adbUsbSlaveAdbHandle)
			&& AdbSlaveReady(adbUsbSlaveAdbHandle))
		{
			while(AdbRouteSlaveMessage(adbUsbSlaveAdbHandle, adbUsbGenericHandle, adbUartHandle))
			{
				//vos_delay_msecs(500);
				//errorPrint("Route Message failed!\r\n");
			}
			errorPrint("Route Message failed!\r\n");
		}
		else
		{
			errorPrint("x");
			if(adbUsbSlaveAdbHandle != NULL)
			{
				//errorPrint("Slave handle ok");
			}
			vos_delay_msecs(100);
		}

	}
}



