/**
 * AndropodInterface - high level adb / serial communication
 *
 * @author Bernhard Julian Nischler  <julian@xdevelop.at>
 * @copyright Copyright 2011, Upper Austria University of Applied Sciences & \
 * Julian Nischler & Bernhard Wrndl-Aichriedler
 */


#include "vos.h"
#include "USB.h"
#include "USBSlave.h"
#include "ioctl.h"
#include "string.h"

#include "Definitions.h"
#include "Debug.h"

#include "AdbSlaveDescriptors.h"
#include "AdbSlave.h"
#include "HardwareControl.h"


// endpoint number
#define USBADBSLAVE_ENDPOINT_IN 0x82
#define USBADBSLAVE_ENDPOINT_OUT 0x02
#define USBADBSLAVE_ENDPOINT_IN_PACKET_SIZE 64
#define USBADBSLAVE_ENDPOINT_IN_PACKET_SIZE_CONFIG USBSLAVE_MAX_PACKET_SIZE_64
#define USBADBSLAVE_ENDPOINT_OUT_PACKET_SIZE 64
#define USBADBSLAVE_ENDPOINT_OUT_PACKET_SIZE_CONFIG USBSLAVE_MAX_PACKET_SIZE_64

//IOCTL definitions
#define VOS_IOCTL_USBADBSLAVE_ATTACH		0
#define VOS_IOCTL_USBADBSLAVE_DETACH		1
#define VOS_IOCTL_USBADBSLAVE_IS_READY		2
#define VOS_IOCTL_USBADBSLAVE_IS_ATTACHED	3

typedef struct _usbAdbSlave_context
{
	// usb slave port handle
	VOS_HANDLE handle;

	// non-zero when initialised
	unsigned char attached;

	// non-zero when set report started
	unsigned char ready;

	// SETUP thread
	vos_tcb_t *tcbSetup;

	// endpoint handles
	usbslave_ep_handle_t ctrlEndpointIn;
	usbslave_ep_handle_t ctrlEndpointOut;
	usbslave_ep_handle_t adbBulkEndpointIn;
	usbslave_ep_handle_t adbBulkEndpointOut;

	// buffer mutex
	vos_mutex_t *writeLock;
	vos_mutex_t *readLock;
	// setup buffer
	unsigned char setupBuffer[9];
	//unsigned char ep_in_buffer[8];

} usbAdbSlave_context;


usbAdbSlave_context AdbSlaveCtx;
vos_driver_t AdbSlaveDriver;


void AdbSlaveSetup(usbAdbSlave_context *ctx);
void AdbSlaveIn(usbAdbSlave_context *ctx);

unsigned char AdbSlaveWrite (char *buf,
							unsigned short num_to_write,
							unsigned short *num_written,
							usbAdbSlave_context *ctx);

unsigned char AdbSlaveRead (char *buf,
							unsigned short num_to_read,
							unsigned short *num_read,
							usbAdbSlave_context *ctx);

unsigned char AdbSlaveIoctl (common_ioctl_cb_t *cb, usbAdbSlave_context *ctx);

unsigned char AdbSlaveRequest(usbAdbSlave_context *ctx);

void AdbSlaveDetach(usbAdbSlave_context *ctx);
unsigned char AdbSlaveAttach(VOS_HANDLE handle, usbAdbSlave_context *ctx);


unsigned char AdbSlaveIoctl (common_ioctl_cb_t *cb, usbAdbSlave_context *ctx);

unsigned char AdbSlaveInit(VOS_HANDLE *usbSlave, uint8 usbSlaveId,
	VOS_HANDLE *adbSlaveHandle, uint8 adbSlaveId, vos_mutex_t *writeLock,
	vos_mutex_t *readLock)
{
	common_ioctl_cb_t AdbAttach;

	//Init usb slave port
	usbslave_init(USBSLAVE_PORT_B, usbSlaveId);

	// Set up function pointers for our driver
	AdbSlaveDriver.flags 		= 0;
	AdbSlaveDriver.read 		= AdbSlaveRead;
	AdbSlaveDriver.write 		= AdbSlaveWrite;
	AdbSlaveDriver.ioctl 		= AdbSlaveIoctl;
	AdbSlaveDriver.interrupt 	= (PF_INT) NULL;
	AdbSlaveDriver.open 		= (PF_OPEN) NULL;
	AdbSlaveDriver.close 		= (PF_CLOSE) NULL;

	// OK - register with device manager
	vos_dev_init(adbSlaveId, &AdbSlaveDriver, &AdbSlaveCtx);

	// defaults
	AdbSlaveCtx.attached 	= 0;
	AdbSlaveCtx.ready 		= 0;
	AdbSlaveCtx.readLock	= readLock;
	AdbSlaveCtx.writeLock	= writeLock;

	//Init main thread
	AdbSlaveCtx.tcbSetup 	= vos_create_thread_ex(20, 0x200, AdbSlaveSetup,"AdbSlave Setup", 2, &AdbSlaveCtx);

	//Initialize mutex -> asume mutex init to be done outside
	//vos_init_mutex(AdbSlaveCtx->writeLock,0);

	// open USB Slave port
	*usbSlave = vos_dev_open(usbSlaveId);
	// open Adb Slave
	*adbSlaveHandle = vos_dev_open(adbSlaveId);

	// attach AdbSlave to USB Slave
	AdbAttach.ioctl_code = VOS_IOCTL_USBADBSLAVE_ATTACH;
	AdbAttach.set.data = (char *)*usbSlave;
	vos_dev_ioctl(*adbSlaveHandle,&AdbAttach);

	return UsbAdbSlaveOk;

}

void AdbSlaveReset(VOS_HANDLE usbSlave)
{
	usbslave_ioctl_cb_t iocb;
	// if we get here, we've been disconnected from the host!
	// put our slave back in a known state ready to be re-connected
	iocb.ioctl_code = VOS_IOCTL_USBSLAVE_DISCONNECT;
	iocb.set = (void *) 0;
	vos_dev_ioctl(usbSlave,&iocb);
}

bool AdbSlaveAttached(VOS_HANDLE AdbSlave)
{
	common_ioctl_cb_t AdbCtl;

	AdbCtl.ioctl_code = VOS_IOCTL_USBADBSLAVE_IS_ATTACHED;
	AdbCtl.set.data = (char *)0;
	AdbCtl.get.param = 0;
	vos_dev_ioctl(AdbSlave,&AdbCtl);

	return (AdbCtl.get.param == 1)?true:false;
}

bool AdbSlaveReady(VOS_HANDLE AdbSlave)
{
	common_ioctl_cb_t AdbCtl;

	AdbCtl.ioctl_code = VOS_IOCTL_USBADBSLAVE_IS_READY;
	AdbCtl.set.data = (char *)0;
	AdbCtl.get.param = 0;
	vos_dev_ioctl(AdbSlave,&AdbCtl);

	return (AdbCtl.get.param == 1)?true:false;
}


/**
 * Read from the USB Slave
 * @param	buf			Buffer to read to
 * @param	num_to_read	Number of bytes to read
 * @param	num_read	Number of bytes actually read from the device
 * @param	ctx			USB Slave context object
 */
unsigned char AdbSlaveRead (char *buf,
							unsigned short num_to_read,
							unsigned short *num_read,
							usbAdbSlave_context *ctx)
{
	usbslave_ioctl_cb_t iocb;
	char code = 0;
	unsigned short i	 	= 0;

	if(num_to_read == 0){ return 0; }

	if (ctx->ready == 1)
	{
		for(i = 0; i < num_to_read;)
		{
			iocb.ioctl_code = VOS_IOCTL_USBSLAVE_TRANSFER;
			iocb.handle = ctx->adbBulkEndpointOut;
			iocb.request.setup_or_bulk_transfer.buffer 				= buf + i; 	/* Do some pointer magic */
			iocb.request.setup_or_bulk_transfer.size
				= (i + USBADBSLAVE_ENDPOINT_IN_PACKET_SIZE >= num_to_read)?(num_to_read - i):USBADBSLAVE_ENDPOINT_IN_PACKET_SIZE;
																				/* Allways read an entire packet */
			iocb.request.setup_or_bulk_transfer.bytes_transferred 	= 0; 		/* Reset bytes transfered file */

			vos_lock_mutex(ctx->readLock);
			vos_unlock_mutex(ctx->readLock);

			if((code = vos_dev_ioctl(ctx->handle,&iocb)) != 0)
			{
				errorPrint("ReadFailed:");
				errorPrintInt(code);
				errorPrint("\r\n");
			}
			i += iocb.request.setup_or_bulk_transfer.bytes_transferred;
		}
	}

	if (num_read)
	{
		*num_read = i;
	}

	if(i != num_to_read)
	{
		errorPrint("Read length error!");
		errorPrintInt(i);
		errorPrint(" of ");
		errorPrintInt(num_to_read);
		errorPrint("\r\n");
		return UsbAdbSlaveError;
	}

	return UsbAdbSlaveOk;
}

/**
 * Write to the USB Slave port
 * @param	buf			Buffer to read to
 * @param	num_to_read	Number of bytes to read
 * @param	num_read	Number of bytes actually read from the device
 * @param	ctx			USB Slave context object
 */
unsigned char AdbSlaveWrite (char *buf,
							unsigned short num_to_write,
							unsigned short *num_written,
							usbAdbSlave_context *ctx)
{
	usbslave_ioctl_cb_t iocb;
	char code = 0;
	unsigned short i = 0;

	if (ctx->ready == 1)
	{
		for(i = 0; i < num_to_write; )
		{
			iocb.ioctl_code = VOS_IOCTL_USBSLAVE_TRANSFER;
			iocb.handle = ctx->adbBulkEndpointIn;
			iocb.request.setup_or_bulk_transfer.buffer = buf + i;
			iocb.request.setup_or_bulk_transfer.size = (i + USBADBSLAVE_ENDPOINT_OUT_PACKET_SIZE >= num_to_write)?num_to_write - i:USBADBSLAVE_ENDPOINT_OUT_PACKET_SIZE;
			iocb.request.setup_or_bulk_transfer.bytes_transferred = 0;

			vos_unlock_mutex(ctx->writeLock);
			if((code = vos_dev_ioctl(ctx->handle,&iocb)) != 0)
			{
				vos_unlock_mutex(ctx->writeLock);
				errorPrint("WriteFailed:");
				errorPrintInt(code);
				errorPrint("\r\n");
			}
			else
			{
				vos_unlock_mutex(ctx->writeLock);
			}

			i += iocb.request.setup_or_bulk_transfer.bytes_transferred;
		}
	}

	if (num_written)
	{
		*num_written = i;
	}

	return UsbAdbSlaveOk;
}


/**
 * USB Slave io control functions
 * @param 	cb		ioctl structure with all parameters
 * @param	ctx		USB slave context
 */
unsigned char AdbSlaveIoctl (common_ioctl_cb_t *cb, usbAdbSlave_context *ctx)
{
	unsigned char status = UsbAdbSlaveInvalidParameter;

	switch (cb->ioctl_code)
	{
	case VOS_IOCTL_USBADBSLAVE_ATTACH:
		status = AdbSlaveAttach((VOS_HANDLE)cb->set.data, ctx);
		break;
	case VOS_IOCTL_USBADBSLAVE_DETACH:
		AdbSlaveDetach(ctx);
		status = UsbAdbSlaveOk;
		break;
	case VOS_IOCTL_USBADBSLAVE_IS_ATTACHED:
		cb->get.param = ctx->attached;
		break;
	case VOS_IOCTL_USBADBSLAVE_IS_READY:
		cb->get.param = ctx->ready;
		break;
	default:
		break;
	}

	return status;
}

/**
 * Detach the USB Slave driver
 * @param	ctx 	USB Slave context
 */
void AdbSlaveDetach(usbAdbSlave_context *ctx)
{
	ctx->attached = 0;
	ctx->ready = 0;
	return;
}

/**
 * Attach the USB slave
 */
unsigned char AdbSlaveAttach(VOS_HANDLE handle, usbAdbSlave_context *ctx)
{
	usbslave_ioctl_cb_t iocb;
	unsigned char status = UsbAdbSlaveOk;

	// save usb slave handle
	ctx->handle = handle;

	if (!ctx->attached)
	{
		//Register Control in endpoint
		iocb.ioctl_code = VOS_IOCTL_USBSLAVE_GET_CONTROL_ENDPOINT_HANDLE;
		iocb.ep = USBSLAVE_CONTROL_IN;
		iocb.get = &ctx->ctrlEndpointIn;
		vos_dev_ioctl(ctx->handle,&iocb);

		//Register Control out endpoint
		iocb.ioctl_code = VOS_IOCTL_USBSLAVE_GET_CONTROL_ENDPOINT_HANDLE;
		iocb.ep = USBSLAVE_CONTROL_OUT;
		iocb.get = &ctx->ctrlEndpointOut;
		vos_dev_ioctl(ctx->handle,&iocb);

		//Register Bulk in Endpoint
		iocb.ioctl_code = VOS_IOCTL_USBSLAVE_GET_BULK_IN_ENDPOINT_HANDLE;
		iocb.ep = USBADBSLAVE_ENDPOINT_IN & 0x0F;
		iocb.get = &ctx->adbBulkEndpointIn;
		vos_dev_ioctl(ctx->handle,&iocb);

		//Register Bulk out Endpoint
		iocb.ioctl_code = VOS_IOCTL_USBSLAVE_GET_BULK_OUT_ENDPOINT_HANDLE;
		iocb.ep = USBADBSLAVE_ENDPOINT_OUT;
		iocb.get = &ctx->adbBulkEndpointOut;
		vos_dev_ioctl(ctx->handle,&iocb);

		//Set Bulk in endpoint packet size
		iocb.ioctl_code = VOS_IOCTL_USBSLAVE_SET_ENDPOINT_MAX_PACKET_SIZE;
		iocb.ep = USBADBSLAVE_ENDPOINT_IN & 0x0F;
		iocb.request.ep_max_packet_size  = USBADBSLAVE_ENDPOINT_IN_PACKET_SIZE_CONFIG;
		vos_dev_ioctl(ctx->handle,&iocb);

		//Set bulk out endpoint packet size
		iocb.ioctl_code = VOS_IOCTL_USBSLAVE_SET_ENDPOINT_MAX_PACKET_SIZE;
		iocb.ep = USBADBSLAVE_ENDPOINT_OUT;
		iocb.request.ep_max_packet_size  = USBADBSLAVE_ENDPOINT_OUT_PACKET_SIZE_CONFIG;
		vos_dev_ioctl(ctx->handle,&iocb);

		//Mark as attached
		ctx->attached = 1;
	}

	return status;
}

/////////////////////////////////////////////////// SETUP ////////////////////////////////////////////////////////

void AdbSlaveSetup(usbAdbSlave_context *ctx)
{
	usbslave_ioctl_cb_t iocb;
	usb_deviceRequest_t *devReq;
	unsigned char bmRequestType;

	while (1) {

		iocb.ioctl_code = VOS_IOCTL_USBSLAVE_WAIT_SETUP_RCVD;
		iocb.request.setup_or_bulk_transfer.buffer = ctx->setupBuffer;
		iocb.request.setup_or_bulk_transfer.size = 9;

		vos_dev_ioctl(ctx->handle,&iocb);

		devReq = (usb_deviceRequest_t *)ctx->setupBuffer;
		bmRequestType = devReq->bmRequestType & 0x60;

		if (bmRequestType == USB_BMREQUESTTYPE_STANDARD) {
			AdbSlaveRequest(ctx);
		}

	}
	return;
}


void ack_request(usbAdbSlave_context *ctx)
{
	usbslave_ioctl_cb_t iocb;

	iocb.ioctl_code = VOS_IOCTL_USBSLAVE_SETUP_TRANSFER;
	iocb.handle = ctx->ctrlEndpointIn;
	iocb.request.setup_or_bulk_transfer.buffer = (void *) 0;
	iocb.request.setup_or_bulk_transfer.size = 0;
	vos_dev_ioctl(ctx->handle,&iocb);
}

void set_address_request(usbAdbSlave_context *ctx,unsigned char addr)
{
	usbslave_ioctl_cb_t iocb;

	iocb.ioctl_code = VOS_IOCTL_USBSLAVE_SET_ADDRESS;
	iocb.set = (void *) addr;
	vos_dev_ioctl(ctx->handle,&iocb);

	ack_request(ctx);
}

void set_configuration_request(usbAdbSlave_context *ctx,unsigned char config)
{
	usbslave_ioctl_cb_t iocb;

	iocb.ioctl_code = VOS_IOCTL_USBSLAVE_SET_CONFIGURATION;
	iocb.set = (void *) config;
	vos_dev_ioctl(ctx->handle,&iocb);

	ack_request(ctx);
}

void set_feature_request(usbAdbSlave_context *ctx, usbslave_ep_handle_t ep)
{
	usbslave_ioctl_cb_t iocb;

	iocb.ioctl_code = VOS_IOCTL_USBSLAVE_ENDPOINT_STALL;
	iocb.ep = ep;
	vos_dev_ioctl(ctx->handle,&iocb);

	ack_request(ctx);
}

void clear_feature_request(usbAdbSlave_context *ctx, usbslave_ep_handle_t ep)
{
	usbslave_ioctl_cb_t iocb;

	iocb.ioctl_code = VOS_IOCTL_USBSLAVE_ENDPOINT_CLEAR;
	iocb.ep = ep;
	vos_dev_ioctl(ctx->handle,&iocb);

	ack_request(ctx);
}

void set_control_ep_halt(usbAdbSlave_context *ctx)
{
	usbslave_ioctl_cb_t iocb;

	// Performs a protocol stall on endpoint 0
	// Indicates that a request is unsupported
	iocb.ioctl_code = VOS_IOCTL_USBSLAVE_ENDPOINT_STALL;
	iocb.ep = 0;
	vos_dev_ioctl(ctx->handle,&iocb);
}

void get_descriptor_request(usbAdbSlave_context *ctx)
{
	usbslave_ioctl_cb_t iocb;
	usb_deviceRequest_t *devReq;
	unsigned char hValue;
	unsigned char lValue;
	unsigned short wLength;
	unsigned short siz;
	uint32 ul_siz;
	unsigned char *src;
	unsigned char cond;


	devReq = (usb_deviceRequest_t *)ctx->setupBuffer;

	hValue = devReq->wValue >> 8;
	lValue = devReq->wValue & 0xff;

	wLength = devReq->wLength;



	switch (hValue) {

	case USB_DESCRIPTOR_TYPE_DEVICE:

		ul_siz = (uint32) wLength;
		iocb.ioctl_code = VOS_IOCTL_USBSLAVE_SETUP_TRANSFER;
		iocb.handle = ctx->ctrlEndpointIn;
		iocb.request.setup_or_bulk_transfer.buffer = device_descriptor;
		iocb.request.setup_or_bulk_transfer.size = (int16) ul_siz;
		vos_dev_ioctl(ctx->handle,&iocb);

		return;
		break;

	case USB_DESCRIPTOR_TYPE_CONFIGURATION:

		siz = wLength == 9 ? 9 : sizeof(config_descriptor);
		ul_siz = (uint32) siz;
		if ((wLength != siz) && ((siz % 8) == 0)) {
			ul_siz |= 0x80000000;
		}
		iocb.ioctl_code = VOS_IOCTL_USBSLAVE_SETUP_TRANSFER;
		iocb.handle = ctx->ctrlEndpointIn;
		iocb.request.setup_or_bulk_transfer.buffer = config_descriptor;
		iocb.request.setup_or_bulk_transfer.size = (int16) ul_siz;
		vos_dev_ioctl(ctx->handle,&iocb);

		ctx->ready = 1;

		return;

	case USB_DESCRIPTOR_TYPE_STRING:

		if (lValue == 0) {
			src = str0_descriptor;
			siz = sizeof(str0_descriptor);
		}
		else if (lValue == 1) {
			src = str1_descriptor;
			siz = sizeof(str1_descriptor);
		}
		else if (lValue == 2) {
			src = str2_descriptor;
			siz = sizeof(str2_descriptor);
		}
		else if (lValue == 3) {
			src = str3_descriptor;
			siz = sizeof(str3_descriptor);
		}
		cond = (unsigned char) (wLength != siz);
		if (siz > wLength)
			siz = wLength;
		ul_siz = (uint32) siz;
		if (cond && ((siz % 8) == 0)) {
			ul_siz |= 0x80000000;
		}
		iocb.ioctl_code = VOS_IOCTL_USBSLAVE_SETUP_TRANSFER;
		iocb.handle = ctx->ctrlEndpointIn;
		iocb.request.setup_or_bulk_transfer.buffer = src;
		iocb.request.setup_or_bulk_transfer.size = (int16) ul_siz;
		vos_dev_ioctl(ctx->handle,&iocb);
		return;

	default:
		// respond with Request Error
		return;
	}
}


unsigned char AdbSlaveRequest(usbAdbSlave_context *ctx)
{
	usb_deviceRequest_t *devReq;
	unsigned char status = USBSLAVE_OK;
	unsigned char bReq;

	devReq = (usb_deviceRequest_t *)ctx->setupBuffer;
	bReq = devReq->bRequest;

	switch (bReq) {

	case USB_REQUEST_CODE_SET_ADDRESS:
		set_address_request(ctx,devReq->wValue&0xff);
		break;

	case USB_REQUEST_CODE_GET_DESCRIPTOR:
		get_descriptor_request(ctx);
		break;

	case USB_REQUEST_CODE_SET_CONFIGURATION:
		set_configuration_request(ctx,devReq->wValue&0xff);
		break;

	case USB_REQUEST_CODE_CLEAR_FEATURE:
		clear_feature_request(ctx, devReq->wIndex >> 8);
		break;

	case USB_REQUEST_CODE_SET_FEATURE:
		set_feature_request(ctx, devReq->wIndex >> 8);
		break;

	default:
		// force a protocol stall
		set_control_ep_halt(ctx);
		break;
	}

	return status;
}
