/**
 * AndropodInterface - basic hardware control functions
 *
 * @author Bernhard Wrndl-Aichriedler <bwa@xdevelop.at>, Julian Nischler  <julian@xdevelop.at>
 * @copyright Copyright 2011, Upper Austria University of Applied Sciences & \
   Julian Nischler & Bernhard Wrndl-Aichriedler
 */

#include "HardwareControl.h"
#include "Definitions.h"
#include "DigitalIO.h"


bool hardwareControlOvercurrent = false;
uint hardwareControlSupervisorTimeout = 0;
vos_tcb_t *hardwareControlTcb;


/**
 * Initialize the hardware control module
 * @param gpioHandle	Pointer to the GPIO handle
 */
void hardwareControlInit()
{
	SetIoMode(1, 1); //Pin A 1 as output -> LED1
	SetIoMode(2, 1); //Pin A 2 as output -> LED2
	SetIoMode(3, 1); //Pin A 7 as output -> USB FAULT
	SetIoMode(5, 0); //Pin A 6 as input -> VBUS
	SetIoMode(6, 0); //Pin A 6 as input -> !FAULT
	SetIoMode(7, 1); //Pin A 7 as output -> USB FAULT

	hardwareControlOvercurrent = false;

	setStatusLed(true);
	setActivityLed(false);

	setCurrentSwitch(true);

	hardwareControlTcb 	= vos_create_thread_ex(20, 0x80, overcurrentSupervisorThread, "Hardware Supervisor",0);
}

/**
 * Read the USB Host power switch overcurrent signal @ IOBUS10
 */
bool getOvercurrentStatus()
{
	return GetIoValue(IOMUX_OUT_GPIO_PORT_A_6)?false:true;
}

/**
 * Enable the USB Host power switch @ IOBUS11
 */
void setCurrentSwitch(bool enabled)
{
	SetIoValue(IOMUX_OUT_GPIO_PORT_A_7, enabled);
}

/**
 * Overcurrent supervising thread used to monitor the overcurrent
 * condition for the USB Host port.
 * If an overcurrent condition is raised, the power will be cut of, and
 * reenabled after HARDWARE_CONTROL_OVERCURRENT_TIMEOUT loops
 */
void overcurrentSupervisorThread()
{
	/* delay a sec, and wait for everything to set up*/
	vos_delay_msecs(1000);

	while(true)
	{
		/* Delay for 100ms to mimize system load */
		vos_delay_msecs(100);

		/* If overcurrent hasn't been detected before */
		if(!hardwareControlOvercurrent)
		{
			if(getOvercurrentStatus())
			{
				/* Turn off overcurrent switch */
				hardwareControlOvercurrent = true;
				setCurrentSwitch(false);
			}
			else
			{
				/* Turn on overcurrent switch - speed-up not necesarry */
				//hardwareControlOvercurrent = false;
				//setCurrentSwitch(true);
			}
		}
		/* Overcurrent has previously been detected - wait timeout than enable again */
		else
		{
			hardwareControlSupervisorTimeout++;

			if(hardwareControlSupervisorTimeout >= HARDWARE_CONTROL_OVERCURRENT_TIMEOUT)
			{
				/* Turn on overcurrent switch */
				hardwareControlOvercurrent = false;
				setCurrentSwitch(true);
				setStatusLed(true);

				/* Reset timeout counter */
				hardwareControlSupervisorTimeout = 0;
			}
			else
			{
				/* Increment timout counter */
				setStatusLed((hardwareControlSupervisorTimeout/5)%2);
			}
		}
	}
}

/**
 * Initialize the watchdog timer of the VNC32
 */
void watchdogInit()
{
	/* Watchdog is initialized to 11.185s @ 48MHz -> see "wdt calculation.xml" */
	//vos_wdt_enable(29);
}

/**
 * Do a watchdog reset
 */
void watchdogReset()
{
	//vos_wdt_clear();
}

void systemReset()
{
	vos_reset_vnc2();
}

/**
 * Check if AndropodInterface's device port is plugged to a host
 * this is done by simple voltage sensing on the VBUS pin @ IOBUS9
 */
bool pluggedToUsbHost()
{
	return GetIoValue(IOMUX_OUT_GPIO_PORT_A_5);
}

/**
 * Set the status of the status led @ IOBUS2
 * @param enabled	LED enabled
 */
void setStatusLed(bool enabled)
{
	SetIoValue(IOMUX_OUT_GPIO_PORT_A_1, enabled);
}

/**
 * Set the status of the activity led @ IOBUS1
 * @param enabled	LED enabled
 */
void setActivityLed(bool enabled)
{
	SetIoValue(IOMUX_OUT_GPIO_PORT_A_2, enabled);
}

/**
 * Read the status of the jumper
 */
bool getJumperStatus()
{
	return GetIoValue(IOMUX_OUT_GPIO_PORT_A_3)?false:true;
}
