/**
 * AndropodInterface - low level adb communication
 *
 * @author Julian Nischler  <julian@xdevelop.at>, Bernhard Wrndl-Aichriedler <bwa@xdevelop.at>
 * @copyright Copyright 2011, Upper Austria University of Applied Sciences & \
 * Julian Nischler & Bernhard Wrndl-Aichriedler
 */

#include "adb.h"

#include "vos.h"
#include "string.h"
#include "stdlib.h"
#include "Debug.h"

 /* Private prototypes */
bool AdbSendMessage(VOS_HANDLE handle, amessage *m);
bool AdbSendData(VOS_HANDLE handle, char *data, uint16 length);
bool AdbSendMessageWithString(VOS_HANDLE handle, amessage *m, char *string);
bool AdbReadMessage(VOS_HANDLE handle, amessage *m);
bool AdbRedirectData(VOS_HANDLE inHandle, VOS_HANDLE outHandle, uint16 length);
bool AdbCheckHeader(amessage *m);
void AdbSetDataCheck(amessage *m, char *data);
void AdbSetDataLength(amessage *m, char *data);
void AdbResetMessage(amessage *m);

/**
 * Id of the connected remote id
 */
unsigned int adbRemoteId = 0;

/**
 * Slave is connected
 */
bool adbConnected 	= false;

/**
 * Stream to slave is open
 */
bool adbOpen 		= false;

/**
 * Mutex used for read locking
 */
vos_mutex_t adbReadWriteMutex;

/**
 * Temporary message stored globally,
 * as there are problems with the malloc function
 * this might consume up to 1/4 of the system memory
 */
char AdbTempData[MAX_PAYLOAD];
/**
 * Slave is connected
 */
bool adbSlaveConnected 	= false;


/**
 * Initialize the adb mutexes
 */
void AdbInit()
{
	vos_init_mutex(&adbReadWriteMutex, 0);
}

/**
 * Check if slave device is connected
 * @return 	Slave connected
 */
bool AdbIsConnected()
{
	return adbConnected;
}

/**
 * Check if connection port to slave device is open
 * @return	Stream is open
 */
bool AdbIsOpen()
{
	return adbOpen;
}

/**
 * Receive message from the slave and respond to the given packet
 * if outHandle is given, special data packets received will be
 * forwared to this handle
 * @param	inHandle		USB input handle
 * @param	outHandle		UART output handle
 * @param	unhandledOut	USB slave handle
 * @return	Message successfully routed
 */
bool AdbRouteMessage(VOS_HANDLE inHandle, VOS_HANDLE outHandle, VOS_HANDLE unhandledOut)
{
	bool retvalue = false;
	bool sendReady = false;
	char data[5];

	VOS_HANDLE outputHandle = NULL;
	VOS_HANDLE packageHandle = NULL;

	amessage msg;
	AdbResetMessage(&msg);

	//Try reading data
	if(AdbReadMessage(inHandle, &msg) != true)
	{
		adbConnected = false;
		adbOpen = false;
		retvalue = false;
	}
	else //Read successfull
	{
		if(msg.command == A_CNXN) 			//Connect command received
		{
			adbConnected = true;
			retvalue = true;
		}
		else if(msg.command == A_SYNC)		//Sync command received
		{
			//we should and we will never receive sync packets
			adbConnected = false;
			retvalue = false;
		}
		else if(msg.command == A_OKAY && msg.arg1 == adbLocalId)
		{  									//If we get a okay command
			if(adbConnected == true)
			{
				adbRemoteId = msg.arg0;
				adbOpen = true;
			}
			retvalue = true;
		}
		else if(msg.command == A_CLSE)
		{									//If we get a close command
			if(adbOpen == true)				//Open connection is closed
			{
				adbConnected = false;
				adbOpen = false;
				adbRemoteId = 0;
				retvalue = false;
			}
			else
			{
				adbOpen = false;
				adbRemoteId = 0;
				retvalue = true;
			}
		}
		else if(msg.command == A_WRTE && msg.arg1 == adbLocalId)
		{										//If we get a write command
			outputHandle = outHandle;			//Redirect the stream
			sendReady = true;
		}
		else								//All other commands
		{
			outputHandle = unhandledOut;
			packageHandle = unhandledOut;
			retvalue = true;
		}

		if(packageHandle != NULL)
		{
			AdbSendMessage(packageHandle, &msg);
		}

		AdbRedirectData(inHandle, outputHandle, msg.data_length);

		if(outputHandle)
		{
			errorPrint("An:");
			errorPrintAdbMessage(&msg);
		}

/*
		if(msg.command == A_WRTE)
		{
			errorPrint("Reading data:");
			AdbReadMessage(inHandle, &msg);
			errorPrintAdbMessage(&msg);
			errorPrint("Reading data done");
		}
*/
	}

	if(sendReady)
	{
		retvalue = AdbSendReady(inHandle);  //Ack data
	}

	return retvalue;
}

/**
 * Receive message from the slave and respond to the given packet
 * if outHandle is given, special data packets received will be
 * forwared to this handle
 * @param	inHandle		USB slave handle
 * @param	outHandle		USB host output handle
 * @param	unhandledOut	UART output handle
 * @return	Message successfully routed
 */
bool AdbRouteSlaveMessage(VOS_HANDLE inHandle, VOS_HANDLE outHandle, VOS_HANDLE unhandledOut)
{
	bool retvalue = true;
	bool redirectMessage 	= false;
	bool replyMessage		= false;
	char data[5];

	char replyMessageData[30];
	VOS_HANDLE outputHandle = NULL;

	amessage msg;
	amessage replyMsg;

	AdbResetMessage(&msg);
	AdbResetMessage(&replyMsg);

	errorPrint("PWait\r\n");

	//Try reading data
	if(AdbReadMessage(inHandle, &msg) != true)
	{
		errorPrint("Error\n\r");

		adbSlaveConnected 	= false;
		retvalue 			= false;
	}
	else //Read successfull
	{
		if(msg.command == A_CNXN) 			//Connect command received
		{
			//errorPrint("Pc:Connect\r\n");

			outputHandle = NULL;

			replyMessage = true;
			replyMsg.command		= A_CNXN;
			replyMsg.arg0			= A_VERSION;
			replyMsg.arg1			= MAX_PAYLOAD;

			memcpy(replyMessageData, adbSlaveConnectBanner, strlen(adbSlaveConnectBanner));

			adbSlaveConnected = true;
		}
		else if(msg.command == A_CLSE)	//If we get a close command, and ignore it
		{
			outputHandle = NULL;

			adbSlaveConnected = false;
		}
		else if(AdbIsConnected())				//All other commands, if connected
		{
			redirectMessage = true;
			outputHandle = outHandle;
		}
		else								//All request with no devices connected should be closed
		{
			outputHandle = NULL;
			replyMessage = true;
			replyMsg.command = A_CLSE;
			replyMsg.arg0 = 0;
			replyMsg.arg1 = msg.arg0;  // close all open commands

			replyMessageData[0] = 0;
		}

		if(redirectMessage)
		{
			AdbSendMessage(outputHandle,&msg);
		}

		AdbRedirectData(inHandle, outputHandle, msg.data_length);

		if(replyMessage)
		{
			AdbSendMessageWithString(inHandle, &replyMsg, replyMessageData);
		}

		/*
		if(msg.command == A_WRTE)
		{
			replyMsg.command = A_OKAY;
			replyMsg.arg1 = msg.arg0;
			replyMsg.arg0 = msg.arg1;
			replyMsg.data_length = 0;
			AdbSendMessage(outputHandle, &replyMsg);
		}

		if(redirectMessage)
		{
			errorPrint("Pc:");
			errorPrintAdbMessage(&msg);
		}

		if(replyMessage)
		{
			errorPrint("Pc:");
			errorPrintAdbMessage(&msg);

			errorPrint("If:");
			errorPrintAdbMessage(&replyMsg);
		}
		*/
	}

	return retvalue;
}

/**
 * Read the next message from the slave and integrity check the packet
 * @param 	handle	Device handle
 * @return	read / crc check successfull
 */
bool AdbReadMessage(VOS_HANDLE handle, amessage *m)
{
	unsigned int count = 0;

	if(vos_dev_read(handle, (void *)m, sizeof(amessage), &count) != 0) {
		errorPrint("read-msg: usbF\r\n");
		return false;
	}
	if(AdbCheckHeader(m) == false) {
		errorPrint("read-msg: headerF\r\n");

		errorPrintAdbData(m, count);
		errorPrintInt(count);

		return false;
	}

	return true;
}

/**
 * Redirect a given amout of bytes from the one stream to another
 * @param 	inHandle	Input device handle - USB
 * @param	outHandle	Output device handle - UART
 * @param	length		Byte count to redirect
 * @return	operation successfull
 */
bool AdbRedirectData(VOS_HANDLE inHandle, VOS_HANDLE outHandle, uint16 length)
{
	bool result = true;
	unsigned short rlength = 0;
	int i = 0;

	if(length == 0){return true;}
	if(length > MAX_PAYLOAD){return false;}

	if(vos_dev_read(inHandle, (void*)AdbTempData, length, &rlength) == 0)
	{
		if(outHandle != NULL)
		{
			if(vos_dev_write(outHandle, (void*)AdbTempData, rlength, NULL) != 0)
			{
				errorPrint("Write Error\r\n");
				result = false;
			}
		}
	}
	else
	{
		errorPrint("Read error\r\n");
		result = false;
	}
/*
	AdbTempData[rlength] = 0;
	errorPrint(AdbTempData);

	for(i = 0; i < rlength; i++)
	{
		errorPrintInt(AdbTempData[i]);
		errorPrint("\r\n");
	}
*/
	if(rlength != length)
	{
		errorPrint("AdbRedirectData read length error");
		errorPrintInt(length);
		errorPrint(" of ");
		errorPrintInt(rlength);
		errorPrint("\r\n");
	}


	return result;
}


/**
 * Send a connect command to the slave
 * @param 	handle	Device handle
 */
bool AdbSendConnect(VOS_HANDLE handle)
{
	amessage adbTempMessage;

	bool result = true;
	vos_lock_mutex(&adbReadWriteMutex);

	AdbResetMessage(&adbTempMessage);

	adbConnected = false;
	adbOpen = false;
	adbRemoteId = 0;

	adbTempMessage.command		= A_CNXN;
	adbTempMessage.arg0			= A_VERSION;
	adbTempMessage.arg1			= MAX_PAYLOAD;

	result = AdbSendMessageWithString(handle, &adbTempMessage, adbConnectBanner);

	vos_unlock_mutex(&adbReadWriteMutex);
	return result;
}

/**
 * Send a open command to the slave
 * @param 	handle	Device handle
 */
bool AdbSendOpen(VOS_HANDLE handle)
{
	amessage adbTempMessage;

	bool result = true;
	vos_lock_mutex(&adbReadWriteMutex);

	AdbResetMessage(&adbTempMessage);

	adbOpen = false;
	adbRemoteId = 0;

	adbTempMessage.command = A_OPEN;
	adbTempMessage.arg0 = adbLocalId;
	adbTempMessage.arg1 = 0;

	result = AdbSendMessageWithString(handle, &adbTempMessage, adbRemoteDest);

	vos_unlock_mutex(&adbReadWriteMutex);

	return result;
}

/**
 * Send a ready command to the slave
 * @param 	handle	Device handle
 */
bool AdbSendReady(VOS_HANDLE handle)
{
	amessage adbTempMessage;
	bool result = true;
	vos_lock_mutex(&adbReadWriteMutex);

	AdbResetMessage(&adbTempMessage);

	adbTempMessage.command = A_OKAY;
	adbTempMessage.arg0 = adbLocalId;
	adbTempMessage.arg1 = adbRemoteId;

	result = AdbSendMessage(handle, &adbTempMessage);

	vos_unlock_mutex(&adbReadWriteMutex);
	return result;
}

/**
 * Send a sync command to the slave
 * @param 	handle	Device handle
 * @param	online	Data array to write
 */
bool AdbSendSync(VOS_HANDLE handle,uint8 online)
{
	amessage adbTempMessage;
	bool result = true;

	vos_lock_mutex(&adbReadWriteMutex);

	AdbResetMessage(&adbTempMessage);

	adbTempMessage.command = A_SYNC;
	adbTempMessage.arg0 = (online == 1) ? 1 : 0;
	adbTempMessage.arg1 = (online == 1) ? 1 : 0;

	result = AdbSendMessage(handle, &adbTempMessage);

	vos_unlock_mutex(&adbReadWriteMutex);
	return result;
}

/**
 * Send a write command with the given payload to the slave
 * @param 	handle	Device handle
 * @param	data	Data array to write
 * @param	length	Lenght of the data to transmit
 */
bool AdbSendWrite(VOS_HANDLE handle, char* data, uint16 length)
{
	amessage adbTempMessage;
	bool result = true;

	vos_lock_mutex(&adbReadWriteMutex);

	AdbResetMessage(&adbTempMessage);

	adbTempMessage.command = A_WRTE;
	adbTempMessage.arg0 = adbLocalId;
	adbTempMessage.arg1 = adbRemoteId;

	adbTempMessage.data_length = length;
	AdbSetDataCheck(&adbTempMessage, data);

	result = false;

	if(AdbSendMessage(handle, &adbTempMessage) == true)
	{
		if(AdbSendData(handle, data, length) == true)
		{
			result = true;
		}
	}

	vos_unlock_mutex(&adbReadWriteMutex);
	return result;
}

/**
 * Send a sync command to the slave
 * @param 	handle	Device handle
 * @return	Operation successfull
 */
bool AdbSendClose(VOS_HANDLE handle)
{
	amessage adbTempMessage;
	bool result = 0;
	vos_lock_mutex(&adbReadWriteMutex);

	AdbResetMessage(&adbTempMessage);

	adbConnected = false;
	adbOpen = false;
	adbRemoteId = 0;

	adbTempMessage.command = A_CLSE;
	adbTempMessage.arg0 = adbLocalId;
	adbTempMessage.arg1 = adbRemoteId;

	result = AdbSendMessage(handle, &adbTempMessage);

	vos_unlock_mutex(&adbReadWriteMutex);
	return result;
}

/**
 * Send an adb message with a string as payload
 * @param 	handle	Device handle
 * @return	Operation successfull
 */
bool AdbSendMessageWithString(VOS_HANDLE handle, amessage *m, char *string)
{
	AdbSetDataLength(m, string);
	AdbSetDataCheck(m, string);

	if(AdbSendMessage(handle,m) == true)
	{
		return AdbSendData(handle, string, m->data_length);
	}

	return false;
}

/**
 * Send an adb message
 * @param 	handle	Device handle
 * @param 	m		Message to send
 * @return	Operation successfull
 */
bool AdbSendMessage(VOS_HANDLE handle, amessage *m)
{
	m->magic = m->command ^ 0xffffffff;

	debugPrint("Send Message: \r\n");
	debugPrintAdbMessage(m);

	if(vos_dev_write(handle, (void*)m, sizeof(amessage), NULL) != 0) {
		errorPrint("send-msg: usbF\r\n");
		return false;
	}
	return true;
}

/**
 * Send an adb data
 * @param 	handle	Device handle
 * @param 	data	Data to send
 * @param	length	Length of the data
 * @return	Operation successfull
 */
bool AdbSendData(VOS_HANDLE handle, char *data, uint16 length)
{
	if(length > 0)
	{
		debugPrint("Send Data: \r\n");
		debugPrintAdbData(data,length);

		if(vos_dev_write(handle, (void*)data, length, NULL) != 0) {
			errorPrint("send-data: usbF\r\n");
			return false;
		}
		return true;
	}
	return false;
}

/**
 * Check header for error with the maxlength and the magic number
 * @param 	m		Message to check
 */
bool AdbCheckHeader(amessage *m)
{
	if(m->magic != (m->command ^ 0xffffffff)) {
		//errorPrint("AdbCheckHeader: invalid magic\r\n");
		return false;
	}

	if(m->data_length > MAX_PAYLOAD) {
		//errorPrint("data exceeds maximum payload \r\n");
		return false;
	}

	return true;
}

/**
 * Set the length of the data to the strlen of the given
 * @param 	m		Message to check
 * @param	data	Payload data
 */
void AdbSetDataLength(amessage *m, char *data)
{
	m->data_length = strlen((char*) data) + 1;
}

/**
 * Create the checksum for the payload data
 * @param 	m		Message to create the checksum
 * @param 	data	Payload data
 */
void AdbSetDataCheck(amessage *m, char *data)
{
	unsigned int sum;
	unsigned int count;
	unsigned char *x;

	count = m->data_length;
	x = (unsigned char *) data;
	sum = 0;
	while(count-- > 0){
		sum += *x++;
	}
	m->data_check = sum;
}

/**
 * Reset the given message
 * @param 	m	Message to reset
 */
void AdbResetMessage(amessage *m)
{
	m->command = 0;
	m->arg0 = 0;
	m->arg1 = 0;
	m->data_length = 0;
	m->data_check = 0;
	m->magic = 0;
}


