/**
 * AndropodInterface - low level adb communication
 *
 * @author Julian Nischler  <julian@xdevelop.at>, Bernhard Wrndl-Aichriedler <bwa@xdevelop.at>
 * @copyright Copyright 2011, Upper Austria University of Applied Sciences & \
 * Julian Nischler & Bernhard Wrndl-Aichriedler
 */

#ifndef _ADB_H_
#define _ADB_H_ 1

#include "Definitions.h"

/**
 * Device class of a android devices' debug interface
 */
#define USB_CLASS_ANDROID                   0xff

/**
 * Subclass of the android debug interface
 */
#define USB_SUBCLASS_ANDROID_ADB            0x42

/**
 * Protocol id of the android debug interface
 */
#define USB_PROTOCOL_ANDROID_DEFAULT        0x01


/**
 * Maximum payload size in bytes
 * this directly reflects the memory usage of the driver
 */
#define MAX_PAYLOAD 2048//4096

/**
 * Definition of the default packet commands
 */
#define A_SYNC 0x434e5953
#define A_CNXN 0x4e584e43
#define A_OPEN 0x4e45504f
#define A_OKAY 0x59414b4f
#define A_CLSE 0x45534c43
#define A_WRTE 0x45545257

/**
 * Version of the Adv Protocol
 */
#define A_VERSION 0x01000000        // ADB protocol version

/**
 * Adb connection string
 * this is how the host reports to the device
 */
#define adbConnectBanner 		"host:0815:AndropodInterface"
#define adbSlaveConnectBanner 	"device:0815:AndropodInterface"

/**
 * Local connection id (only one connection allowed)
 */
#define adbLocalId  1331

 /**
  * Destination string passed to the connect
  * "shell:" - for shell access
  * "tcp:<port>" - for a tcp connection to a certain port
  */
#define adbRemoteDest  "tcp:1337"
//#define adbRemoteDest  "shell:"

/**
 * Definition of the adb transfer structure
 */
typedef struct
{
	unsigned int command;       /* command identifier constant      */
	unsigned int arg0;          /* first argument                   */
	unsigned int arg1;          /* second argument                  */
	unsigned int data_length;   /* length of payload (0 is allowed) */
	unsigned int data_check;    /* checksum of data payload         */
	unsigned int magic;         /* command ^ 0xffffffff             */
} amessage;

/**
 * Initialize the adb mutexes
 */
void AdbInit();

/**
 * Check if slave device is connected
 * @return 	Slave connected
 */
bool AdbIsConnected();

/**
 * Check if connection port to slave device is open
 * @return	Stream is open
 */
bool AdbIsOpen();

/**
 * Send a connect command to the slave
 * @param 	handle	Device handle
 */
bool AdbSendConnect(VOS_HANDLE handle);

/**
 * Send a open command to the slave
 * @param 	handle	Device handle
 */
bool AdbSendOpen(VOS_HANDLE handle);

/**
 * Send a ready command to the slave
 * @param 	handle	Device handle
 */
bool AdbSendReady(VOS_HANDLE handle);

/**
 * Send a sync command to the slave
 * @param 	handle	Device handle
 * @param	online	Data array to write
 */
bool AdbSendSync(VOS_HANDLE handle,uint8 online);

/**
 * Send a write command with the given payload to the slave
 * @param 	handle	Device handle
 * @param	data	Data array to write
 * @param	length	Lenght of the data to transmit
 */
bool AdbSendWrite(VOS_HANDLE handle, char* data, uint16 length);

/**
 * Send a sync command to the slave
 * @param 	handle	Device handle
 * @return	Operation successfull
 */
bool AdbSendClose(VOS_HANDLE handle);

/**
 * Receive message from the slave and respond to the given packet
 * if outHandle is given, special data packets received will be
 * forwared to this handle
 * @param	inHandle		USB input handle
 * @param	outHandle		UART output handle
 * @param	unhandledOut	USB slave handle
 * @return	Message successfully routed
 */
bool AdbRouteMessage(VOS_HANDLE inHandle, VOS_HANDLE outHandle, VOS_HANDLE unhandledOut);

/**
 * Receive message from the slave and respond to the given packet
 * if outHandle is given, special data packets received will be
 * forwared to this handle
 * @param	inHandle		USB input handle
 * @param	outHandle		UART output handle
 * @param	unhandledOut	USB slave handle
 * @return	Message successfully routed
 */
bool AdbRouteSlaveMessage(VOS_HANDLE inHandle, VOS_HANDLE outHandle, VOS_HANDLE unhandledOut);


#endif
